//	Image Noise Removal Filter
//	Copyright (C) 2002 Karlis Freivalds
//
//	This program is free software; you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation; either version 2 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program; if not, write to the Free Software
//	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

#include <math.h>
#include <stdlib.h>
#include <string.h>
#include "reduce.h"
#include "colorDenoise.h"
#include "median.h"

/*
    This method performs the filtering
*/
void colorDenoise::process(int *Y, int *Cr, int *Cb, int width, int height)
{
    this->width = width;
    this->height = height;
    int size = width*height;

    int *fY = new int[size];
    int *fR = new int[size];
    int *fB = new int[size];

    memcpy(fY,Y,size*sizeof(int));
    memcpy(fR,Cr,size*sizeof(int));
    memcpy(fB,Cb,size*sizeof(int));

    medianFilter(fY, fR, fB, width, height);
    createEdgeGraph(fY, fR, fB);

    adaptiveSmoothing s;
    s.setEdgeGraph(edgeGraph);
    s.setParameters(Ry, preserveEdges);
    s.process(Y, width,height);

  if(!this->doReducing)
  {
      s.setParameters(Rb,preserveEdges);
      s.process(Cb, width,height);

      s.setParameters(Rr,preserveEdges);
      s.process(Cr, width,height);

      deleteEdgeGraph(width*height);
  }else
  {
      deleteEdgeGraph(width*height);

      int newWidth = (width+2)/2;
      int newHeight = (height+2)/2;

      int *newY = new int[newWidth*newHeight];
      int *newR = new int[newWidth*newHeight];
      int *newB = new int[newWidth*newHeight];

      shrink2x(Y, newY, width, height);
      shrink2x(Cr, newR, width, height);
      shrink2x(Cb, newB, width, height);


      this->width = newWidth;
      this->height = newHeight;

      createEdgeGraph(newY, newR, newB);
      s.setEdgeGraph(edgeGraph);

      s.setParameters(Rb/2,preserveEdges);
      s.process(newB, newWidth,newHeight);

      s.setParameters(Rr/2,preserveEdges);
      s.process(newR, newWidth,newHeight);

      unShrink2x(Cr, newR, width, height);
      unShrink2x(Cb, newB, width, height);

      delete [] newY;
      delete [] newR;
      delete [] newB;

      deleteEdgeGraph(newWidth*newHeight);
  }

  delete [] fY;
  delete [] fR;
  delete [] fB;
}

/*
    This method sets the parameters: radius for each color plane,
    edge preservation strength, reduction of Cr and Cb planes
*/
void colorDenoise::setParameters(float Ry, float Rr, float Rb, int preserveEdges,bool doReducing)
{
    this->Ry = Ry;
    this->Rb = Rb;
    this->Rr = Rr;
    this->preserveEdges =preserveEdges;
    this->doReducing = doReducing;

}

inline double sqr1(double a)
{
    return a*a;
}

/*
    This method returns the edge weight based on the color differences between
    pixels
*/
int colorDenoise::getEdgeWeight(int dy, int dr, int db)
{
    return sqrt(sqr1(dy*wY)+sqr1(dr*wR)+sqr1(db*wB))+0.5;
}

void colorDenoise::deleteEdgeGraph(int length)
{
    for(int i=0;i<length;i++)
    {
        delete [] edgeGraph[i];
    }

    delete [] edgeGraph;
}

/*
    This method creates the pixel graph
*/
void colorDenoise::createEdgeGraph(int *Y, int *Cr, int *Cb)
{
    edgeGraph = new int*[width*height];

    if(Ry>=0.1)
        wY = 1.0/sqr1(Ry);
    else wY = 1;

    if(Rr>=0.1)
        wR = 1.0/sqr1(Rr);
    else wR = 0;

    if(Rb>=0.1)
        wB = 1.0/sqr1(Rb);
    else wB = 0;

    invWeight = 1.0/sqrt(sqr1(wY)+ sqr1(wR)+sqr1(wB));
    wY*=invWeight;
    wR*=invWeight;
    wB*=invWeight;

    int counter = 0;
    for(int k=0;k<height;k++)
    for(int i=0;i<width;i++)
    {
        int ind = i+k*width;
        int startY = Y[ind];
        int startR = Cr[ind];
        int startB = Cb[ind];

        edgeGraph[counter] = new int[8];
        int *graphElement = edgeGraph[counter];
        memset(graphElement, 0, 8*sizeof(int));

        if(i>0)
        {
            graphElement[0] = getEdgeWeight(startY-Y[ind-1],startR-Cr[ind-1],startB-Cb[ind-1]);
        }

        if(k>0)
        {
           int ind1 = ind-width;
           graphElement[1] = getEdgeWeight(startY-Y[ind1],startR-Cr[ind1],startB-Cb[ind1]);
        }

        if(i<width-1)
        {
            int ind1 = ind+1;
            graphElement[2] = getEdgeWeight(startY-Y[ind1],startR-Cr[ind1],startB-Cb[ind1]);
        }

        if(k<height-1)
        {
            int ind1 = ind+width;
            graphElement[3] = getEdgeWeight(startY-Y[ind1],startR-Cr[ind1],startB-Cb[ind1]);
        }

        /////////////////////
        if(i>0 && k>0)
        {
            int ind1 = ind-width-1;
            graphElement[4] = getEdgeWeight(startY-Y[ind1],startR-Cr[ind1],startB-Cb[ind1]);
        }

        if(i>0 && k<height-1)
        {
            int ind1 = ind-1+width;
            graphElement[5] = getEdgeWeight(startY-Y[ind1],startR-Cr[ind1],startB-Cb[ind1]);
        }

        if(i<width-1 && k<height-1)
        {
            int ind1 = ind+1+width;
            graphElement[6] = getEdgeWeight(startY-Y[ind1],startR-Cr[ind1],startB-Cb[ind1]);
        }

        if(i<width-1 && k>0)
        {
            int ind1 = ind+1-width;
            graphElement[7] = getEdgeWeight(startY-Y[ind1],startR-Cr[ind1],startB-Cb[ind1]);
        }

        counter++;
    }
}

