//	Image Noise Removal Filter
//	Copyright (C) 2002 Karlis Freivalds
//
//	This program is free software; you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation; either version 2 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program; if not, write to the Free Software
//	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

#include <string.h>
#include <math.h>
#include <stdlib.h>
#include "queue.h"
#include "smooth.h"


adaptiveSmoothing::adaptiveSmoothing()
{
    distWeight = 12*256;
    maxDistance = 40*256;
    range = 3;
}
void adaptiveSmoothing::setEdgeGraph(int **graph)
{
    this->edgeGraph = graph;
}

/*
    This method sets the radius and edge preservation parameters.
*/
void adaptiveSmoothing::setParameters(float radius, int preserveEdges)
{
    if(preserveEdges<1) preserveEdges = 1;
    maxDistance = radius*(preserveEdges+512);
    distWeight = preserveEdges;
    range = radius+0.999;
    this->radius = radius;
}


/*
    This method performs the filtering
*/
void adaptiveSmoothing::process(int *data, int width, int height)
{
    if(radius<0.1) return;

    this->data = data;
    this->width = width;
    this->height = height;

    maxLength = sqr(2*range+1);

    hist = new point[maxLength];
    seen = new char[maxLength];
    Q = new PriorityQueue (maxLength);
    createFilter();
    createDistMap();

    enhance();

    delete [] hist;
    delete [] seen;
    delete Q;
    delete [] filter;
    delete [] distMap;
}

/*
    This method precalculates the distance function table.
*/
void adaptiveSmoothing::createDistMap()
{
    distMap = new float[maxLength];
    int l = range*2+1;

    for(int i=-range;i<=range;i++)
    for(int k=-range;k<=range;k++)
    {
        distMap[(i+range)+(k+range)*l] =
            distWeight*sqrt(sqr(i)+sqr(k));
    }

}

/*
    This method precalculates the filter table.
*/
void adaptiveSmoothing::createFilter()
{
    int filterLen = range*16;
    float len1 = radius*16;
    filter = new double[filterLen+1];

    for(int i=0;i<filterLen+1;i++)
        filter[i] = exp(-sqr(i)*3/sqr(len1));

    invFilterLen = ((double)filterLen)/maxDistance;
}

/*
    This method finds all neighbour pixels within maxDistance range from the
    initial pixel (x,y). For each pixel the shortest path value is also obteined.
*/
int adaptiveSmoothing::search(int x, int y)
{
    int l = range*2+1;

    int minx = max(x-range, 0);
    int maxx = min(x+range,width-1);
    int miny = max(y-range, 0);
    int maxy = min(y+range,height-1);

    memset(seen, 0,maxLength*sizeof(char));

    QueueElement elem;
    elem.key =0;
    elem.baseKey =0;
    elem.value =x-minx + (y-miny)*l;
    Q->AddElem(elem);
    int count = 0;

    while(Q->NotEmpty() && count<maxLength)
    {
        Q->GetElem(&elem);
        int ind = elem.value;
        int y1 = ind/l+miny;
        int x1 = ind%l+minx;
        float key = elem.baseKey;

        seen[ind] = 2;
        hist[count].x = x1;
        hist[count].y = y1;
        hist[count++].distance = elem.key;
        int dataInd =x1+y1*width;
        int *graphElement = edgeGraph[dataInd];

        if(x1>minx && seen[ind-1]<=1)
        {
            elem.value = ind-1;
            elem.baseKey = key+graphElement[0];
            elem.key = elem.baseKey + distMap[ind-1];
            if(elem.key<maxDistance)
                Q->AddElem(elem);
        }

        if(y1>miny && seen[ind-l]<=1)
        {
            elem.value = ind-l;
            elem.baseKey = key+graphElement[1];
            elem.key = elem.baseKey + distMap[ind-l];
            if(elem.key<maxDistance)
                Q->AddElem(elem);
        }

        if(x1<maxx && seen[ind+1]<=1)
        {
            elem.value = ind+1;
            elem.baseKey = key+graphElement[2];
            elem.key = elem.baseKey + distMap[ind+1];
            if(elem.key<maxDistance)
                Q->AddElem(elem);
        }

        if(y1<maxy && seen[ind+l]<=1)
        {
            elem.value = ind+l;
            elem.baseKey = key+graphElement[3];
            elem.key = elem.baseKey + distMap[ind+l];
            if(elem.key<maxDistance)
                Q->AddElem(elem);
        }

        /////////////////////
        if(x1>minx && y1>miny && seen[ind-1-l]<=1)
        {
            elem.value = ind-1-l;
            elem.baseKey = key+graphElement[4];
            elem.key = elem.baseKey + distMap[ind-1-l];
            if(elem.key<maxDistance)
                Q->AddElem(elem);
        }

        if(x1>minx && y1<maxy && seen[ind-1+l]<=1)
        {
            elem.value = ind-1+l;
            elem.baseKey = key+graphElement[5];
            elem.key = elem.baseKey + distMap[ind-1+l];
            if(elem.key<maxDistance)
                Q->AddElem(elem);
        }

        if(x1<maxx && y1<maxy && seen[ind+1+l]<=1)
        {
            elem.value = ind+1+l;
            elem.baseKey = key+graphElement[6];
            elem.key = elem.baseKey + distMap[ind+1+l];
            if(elem.key<maxDistance)
                Q->AddElem(elem);
        }

        if(x1<maxx && y1>miny && seen[ind+1-l]<=1)
        {
            elem.value = ind+1-l;
            elem.baseKey = key+graphElement[7];
            elem.key = elem.baseKey + distMap[ind+1-l];
            if(elem.key<maxDistance)
                Q->AddElem(elem);
        }

    }
    return count;
}

/*
    This method returns the filtered pixel value at coordinate (x,y).
*/

int adaptiveSmoothing::enhancePixel(int x, int y)
{
    int count = search(x,y);
    int result;

    if(count==0)
    {
        result = data[x+y*width];
    }
    else
    {
        double sum = 0;
        double divisor = 0;

        for(int i=0;i<count;i++)
        {
            double dist = hist[i].distance;
            double weight = filter[(int)(dist*invFilterLen)];

            sum += data[hist[i].x + hist[i].y*width]*weight;
            divisor+=weight;
        }

        result = sum/divisor;
    }

    return result;

}

/*
    This method applies filter for the whole image.
*/
void adaptiveSmoothing::enhance()
{
    int *data1 = new int[width*height];

    for(int k=0;k<height;k++)
    for(int i=0;i<width;i++)
    {
        int newValue = enhancePixel(i,k);
        data1[k*width+i] = newValue;
    }

    memcpy(data,data1,width*height*sizeof(int));

    delete [] data1;
}
